// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file BoundaryManager.h */

#ifndef AMDIS_BOUNDARYMANAGER_H
#define AMDIS_BOUNDARYMANAGER_H

#include <map>

#include "AMDiS_fwd.h"
#include "Boundary.h"
#include "BoundaryCondition.h"

namespace AMDiS {

  /**
   * \ingroup Assembler
   *
   * \brief
   * A BoundaryManager handles a set of boundary conditions and applies
   * this conditions to DOFVectorBase and DOFMatrix objects. Each DOFVectorBase
   * and each DOFMatrix has its own BoundaryManager.
   */
  class BoundaryManager
  {
  public:
    BoundaryManager(const FiniteElemSpace *feSpace);

    BoundaryManager(BoundaryManager &bm);

    ~BoundaryManager();

    /// Adds a local boundary condition to the list of managed conditions.
    void addBoundaryCondition(BoundaryCondition *localBC) 
    {
      BoundaryType type = localBC->getBoundaryType();
      TEST_EXIT(localBCs[type] == NULL)("there is already a condition for this type\n");
      localBCs[type] = localBC;
    }

    void initMatrix(DOFMatrix *matrix);

    void exitMatrix(DOFMatrix *matrix);

    void initVector(DOFVectorBase<double> *vector);

    void exitVector(DOFVectorBase<double> *vector);

    /** \brief
     * Calls DOFVectorBase::fillBoundaryCondition() for each local boundary condition
     * in \ref localBCs.
     */
    void fillBoundaryConditions(ElInfo *elInfo, DOFVectorBase<double> *vec);

    /** \brief
     * Calls DOFMatrix::fillBoundaryCondition() for each local boundary condition
     * in \ref localBCs.
     */
    void fillBoundaryConditions(ElInfo *elInfo, DOFMatrix *mat);

    /** \brief
     * Calls BoundaryCondition::boundResidual() for each boundary condition in 
     * \ref localBCs.
     */
    double boundResidual(ElInfo *elInfo, 
			 DOFMatrix *matrix,
			 const DOFVectorBase<double> *dv);

    inline BoundaryCondition *getBoundaryCondition(BoundaryType type) 
    {
      return localBCs[type];
    }

    const std::map<BoundaryType, BoundaryCondition*>& getBoundaryConditionMap() 
    {
      return localBCs;
    }

    void setBoundaryConditionMap(const std::map<BoundaryType, BoundaryCondition*>& bcs) 
    {
      localBCs = bcs;
    }

  protected:
    /// Map of managed local boundary conditions.
    std::map<BoundaryType, BoundaryCondition*> localBCs;

    /// Temporary thread-safe variable for functions fillBoundaryconditions.
    std::vector<BoundaryType*> localBounds;

    /// Temporary thread-safe variable for functions fillBoundaryconditions.
    std::vector<Vector<DegreeOfFreedom> > dofIndices;

    /** \brief
     * Stores the number of byte that were allocated in the constructor for
     * each localBounds value. Is used to free the memory in the destructor.
     */
    int allocatedMemoryLocalBounds;
  };

}

#endif
