#include "ElInfo2d.h"
#include "BasisFunction.h"
#include "Element.h"
#include "Line.h"
#include "Triangle.h"
#include "Tetrahedron.h"
#include "FiniteElemSpace.h"
#include "Flag.h"
#include "MacroElement.h"
#include "Mesh.h"
#include "Global.h"
#include "FixVec.h"
#include "DOFVector.h"

namespace AMDiS {

  double ElInfo2d::mat_d1_val[3][3] = {{1.0, 0.0, 0.0}, 
				       {0.0, 1.0, 0.0}, 
				       {0.0, 0.0, 1.0}};
  mtl::dense2D<double> ElInfo2d::mat_d1(mat_d1_val);

  double ElInfo2d::mat_d1_left_val[3][3] = {{0.0, 1.0, 0.5}, 
					    {0.0, 0.0, 0.5},
					    {1.0, 0.0, 0.0}};
  mtl::dense2D<double> ElInfo2d::mat_d1_left(mat_d1_left_val);

  double ElInfo2d::mat_d1_right_val[3][3] = {{0.0, 0.0, 0.5}, 
					     {1.0, 0.0, 0.5},
					     {0.0, 1.0, 0.0}};
  mtl::dense2D<double> ElInfo2d::mat_d1_right(mat_d1_right_val);

  ElInfo2d::ElInfo2d(Mesh *aMesh) 
    : ElInfo(aMesh) 
  {
    e1 = new WorldVector<double>;
    e2 = new WorldVector<double>;
    normal = new WorldVector<double>;
  }

  ElInfo2d::~ElInfo2d()
  {
    delete e1;
    delete e2;
    delete normal;
  }

  void ElInfo2d::fillMacroInfo(const MacroElement * mel)
  {
    FUNCNAME("ElInfo::fillMacroInfo()");
 
    macroElement_ = const_cast<MacroElement*>(mel);
    element_ = const_cast<Element*>(mel->getElement());
    parent_ = NULL;
    level = 0;

    if (fillFlag_.isSet(Mesh::FILL_COORDS) || 
	fillFlag_.isSet(Mesh::FILL_DET)    ||
	fillFlag_.isSet(Mesh::FILL_GRD_LAMBDA)) {

      int vertices = mesh_->getGeo(VERTEX);
      for (int i = 0; i < vertices; i++)
	coord_[i] = mel->coord[i];
    }

    int neighbours = mesh_->getGeo(NEIGH);

    if (fillFlag_.isSet(Mesh::FILL_OPP_COORDS) || 
	fillFlag_.isSet(Mesh::FILL_NEIGH)) {

      bool fill_opp_coords = (fillFlag_.isSet(Mesh::FILL_OPP_COORDS));
    
      for (int i = 0; i < neighbours; i++) {
	MacroElement *macroNeighbour = mel->getNeighbour(i);

	if (macroNeighbour) {
	  neighbour_[i] = macroNeighbour->getElement();	  
	  Element *nb = const_cast<Element*>(neighbour_[i]);

	  int edgeNo = oppVertex_[i] = mel->getOppVertex(i);


	  if (nb->getFirstChild() && (edgeNo != 2)) {  

	    // Search for the next neighbour. In many cases, the neighbour element 
	    // may be refinemed in a way, such that there is no new vertex on the 
	    // common edge. This situation is shown in the following picture: 
	    //
	    //               /|\
	    //              / | \
	    //             /  |  \
	    //            /\  |   \
	    //           /  \ |    \ 
	    //          /    \|     \
	    //          -------------
	    //
	    //            nb     el
	    //
	    // Note that we know (because of the last if statement), that the 
	    // neighbour element has children and the common edge is not the 
	    // refinement edge, which has always the number 2, of our element.


	    if (edgeNo == 0) {
	      // The situation is as follows:
	      //
	      //          -------
	      //          \    /|\
	      //           \  / | \
	      //            \/  |  \
	      //             \  |   \
	      //              \ |    \ 
	      //               \|     \
	      //                -------
	      //
	      //            nb     el
	      // That means, the edge 0 of the same level neighbour is the common
	      // edge, i.e., the direct neighbour is the second child of the same
	      // level neighbour.

	      nb = neighbour_[i] = nb->getSecondChild();
	    } else {
	      // The situation is as shown in the picture above. So the next
	      // neighbour is the first child of the same level neighbour element.
	      nb = neighbour_[i] = nb->getFirstChild();
	    }

	    // In both cases the opp vertex number is 2, as one can see in the 
	    // pictures above.
	    oppVertex_[i] = 2;

	    if (fill_opp_coords) {
	      if (nb->isNewCoordSet()) {
		oppCoord_[i] = *(nb->getNewCoord());
	      } else {
		// In both cases, that are shown in the pictures above, the opp
		// vertex of the neighbour edge is the midpoint of the vertex 0
		// and vertex 1 of the same level neighbour element.
		oppCoord_[i] = (macroNeighbour->coord[0] + 
				macroNeighbour->coord[1]) * 0.5;
	      }
	      
	      switch (i) {
	      case 0:
		// The common edge is the edge 0 of this element.

		switch (edgeNo) {
		case 1:
		  neighbourCoord_[i][0] = macroNeighbour->coord[2];
		  neighbourCoord_[i][1] = macroNeighbour->coord[0];
		  break;
		case 0:		  
		  neighbourCoord_[i][0] = macroNeighbour->coord[1];
		  neighbourCoord_[i][1] = macroNeighbour->coord[2];
		  break;
		default:
		  ERROR_EXIT("Should not happen!\n");
		}
	
		neighbourCoord_[i][2] = oppCoord_[i];
		break;
		
	      case 1:
		// The commonedge is the edge 1 of this element.
		switch (edgeNo) {
		case 0:
		  neighbourCoord_[i][0] = macroNeighbour->coord[1];
		  neighbourCoord_[i][1] = macroNeighbour->coord[2];
		  break;
		case 1:
		  neighbourCoord_[i][0] = macroNeighbour->coord[2];
		  neighbourCoord_[i][1] = macroNeighbour->coord[0];
		  break;
		default:
		  ERROR_EXIT("Should not happen!\n");
		}
		
		neighbourCoord_[i][2] = oppCoord_[i];
		break;
		
	      case 2:
		// I've deleted here some code, be I think that this case is not
		// possible. If an error occurs in this line, please check AMDiS
		// revision <= 476 at the same position.
		ERROR_EXIT("Should not happen!\n");
		break;

	      default:
		std::cout << "------------- Error --------------" << std::endl;
		std::cout << "  Neighbour counter = " << i << "\n";
		std::cout << "  Element index     = " << element_->getIndex() << "\n\n";
		for (int j = 0; j < neighbours; j++) {
		  if (mel->getNeighbour(j)) {
		    std::cout << "  Neighbour " << j << ": " 
			      << mel->getNeighbour(j)->getElement()->getIndex() 
			      << std::endl;
		  } else {
		    std::cout << "  Neighbour " << j << ": not existing" << std::endl;
		  }
		  std::cout << "  OppVertex " << j << ": " 
			    << static_cast<int>(mel->getOppVertex(j)) 
			    << std::endl << std::endl;
		}
		ERROR_EXIT("should not happen!\n");
		break;
	      }
	    }
	  } else {

	    // In this case, we know that the common edge is the refinement edge.
	    // This makes everything much more simpler, because we know that the
	    // next neighbour is equal to the samel level neighbour. If the same
	    // level neighbour would be refinement, also this element must to be 
	    // refinement, because they share the refinement edge.

	    if (fill_opp_coords) {
	      oppCoord_[i] = macroNeighbour->coord[edgeNo];
	      neighbourCoord_[i] = macroNeighbour->coord;	      
	    }
	  }
	} else {
	  neighbour_[i] = NULL;
        }
      }
    }
    
    if (fillFlag_.isSet(Mesh::FILL_BOUND)) {   
      for (int i = 0; i < element_->getGeo(BOUNDARY); i++) {
	boundary_[i] = mel->getBoundary(i);
      }

      for (int i = 0; i < element_->getGeo(PROJECTION); i++) {
	projection_[i] = mel->getProjection(i);
      }
    }
  }


  /****************************************************************************/
  /*   fill ElInfo structure for one child of an element   		    */
  /****************************************************************************/

  void ElInfo2d::fillElInfo(int ichild, const ElInfo *elInfoOld)
  {
    FUNCNAME("ElInfo::fillElInfo()");

    Element *elem = elInfoOld->element_;
    Element *nb;

    Flag fill_flag = elInfoOld->fillFlag_;

    TEST_EXIT_DBG(elem->getFirstChild())("no children?\n");
    element_ = const_cast<Element*>((ichild == 0) ? 
				    elem->getFirstChild() : 
				    elem->getSecondChild());
    TEST_EXIT_DBG(element_)("missing child %d?\n", ichild);

    macroElement_  = elInfoOld->macroElement_;
    fillFlag_ = fill_flag;
    parent_ = elem;
    level = elInfoOld->level + 1;
    iChild = ichild;

    if (fillFlag_.isSet(Mesh::FILL_COORDS) || 
	fillFlag_.isSet(Mesh::FILL_DET)    ||
	fillFlag_.isSet(Mesh::FILL_GRD_LAMBDA)) {
      
      if (elem->isNewCoordSet()) {
	coord_[2] = *(elem->getNewCoord());
      } else {
	coord_[2].setMidpoint(elInfoOld->coord_[0], elInfoOld->coord_[1]);
      }
      
      if (ichild == 0) {
	coord_[0] = elInfoOld->coord_[2];
	coord_[1] = elInfoOld->coord_[0];
      } else {
	coord_[0] = elInfoOld->coord_[1];
	coord_[1] = elInfoOld->coord_[2];
      }
    }

    bool fill_opp_coords = (fill_flag.isSet(Mesh::FILL_OPP_COORDS));

    if (fill_flag.isSet(Mesh::FILL_NEIGH) || fill_opp_coords) {     
      if (ichild == 0) {
	// Calculation of the neighbour 2, its oppCoords and the
	// cooresponding oppVertex.

	neighbour_[2] = elInfoOld->neighbour_[1];
	oppVertex_[2] = elInfoOld->oppVertex_[1];
	
	if (neighbour_[2] && fill_opp_coords) {
	  oppCoord_[2] = elInfoOld->oppCoord_[1];
	  neighbourCoord_[2] = elInfoOld->neighbourCoord_[1];
	}
	
	
	// Calculation of the neighbour 1, its oppCoords and the
	// cooresponding oppVertex.
	
	if (elem->getFirstChild()  &&  
	    elem->getSecondChild()->getFirstChild()  &&  
	    elem->getSecondChild()->getFirstChild()) {
	  
	  neighbour_[1] = elem->getSecondChild()->getSecondChild();
	  oppVertex_[1] = 2;
	  
	  if (fill_opp_coords) {
            if (elem->getSecondChild()->isNewCoordSet()) {
	      oppCoord_[1] = *(elem->getSecondChild()->getNewCoord());
	    } else {      
	      oppCoord_[1].setMidpoint(elInfoOld->coord_[1], 
				       elInfoOld->coord_[2]);
	    }

	    neighbourCoord_[1][0] = coord_[0];
	    neighbourCoord_[1][1] = coord_[2];
	    neighbourCoord_[1][2] = oppCoord_[1];  
	  }
	} else {
	  neighbour_[1] = elem->getSecondChild();
	  oppVertex_[1] = 0;

	  if (fill_opp_coords) {
	    oppCoord_[1] = elInfoOld->coord_[1];

	    neighbourCoord_[1][0] = elInfoOld->coord_[1];
	    neighbourCoord_[1][1] = elInfoOld->coord_[2];
	    neighbourCoord_[1][2] = coord_[2];
	  }
	}


	// Calculation of the neighbour 0, its oppCoords and the
	// cooresponding oppVertex.
	
	nb = elInfoOld->neighbour_[2];
	if (nb) {
	  TEST(elInfoOld->oppVertex_[2] == 2)("invalid neighbour\n"); 
	  TEST_EXIT_DBG(nb->getFirstChild())("missing first child?\n");
	  TEST_EXIT_DBG(nb->getSecondChild())("missing second child?\n");
	 
	  nb = nb->getSecondChild();

	  if (nb->getFirstChild()) {
	    oppVertex_[0] = 2;

	    if (fill_opp_coords) {
	      if (nb->isNewCoordSet()) {
		oppCoord_[0] = *(nb->getNewCoord());
	      } else {
		oppCoord_[0].setMidpoint(elInfoOld->neighbourCoord_[2][1],
					 elInfoOld->neighbourCoord_[2][2]);
	      }

	      neighbourCoord_[0][0].setMidpoint(elInfoOld->neighbourCoord_[2][0],
						elInfoOld->neighbourCoord_[2][1]);
	      neighbourCoord_[0][1] = elInfoOld->neighbourCoord_[2][1];
	      neighbourCoord_[0][2] = oppCoord_[0];
	    }	   
 
	    nb = nb->getFirstChild();
	  } else {
	    oppVertex_[0] = 1;

	    if (fill_opp_coords) {
	      oppCoord_[0] = elInfoOld->oppCoord_[2];    

	      neighbourCoord_[0][0] = elInfoOld->neighbourCoord_[2][0];
	      neighbourCoord_[0][1] = elInfoOld->neighbourCoord_[2][2];
	      neighbourCoord_[0][2].setMidpoint(elInfoOld->neighbourCoord_[2][0],
						elInfoOld->neighbourCoord_[2][1]);
	    }
	  }
	}
	
	neighbour_[0] = nb;
      } else {   /* ichild == 1 */
	// Calculation of the neighbour 2, its oppCoords and the
	// cooresponding oppVertex.

	neighbour_[2] = elInfoOld->neighbour_[0];
	oppVertex_[2] = elInfoOld->oppVertex_[0];

	if (neighbour_[2] && fill_opp_coords) {
	  oppCoord_[2] = elInfoOld->oppCoord_[0];
	  neighbourCoord_[2] = elInfoOld->neighbourCoord_[0];
	}
	

	// Calculation of the neighbour 0, its oppCoords and the
	// cooresponding oppVertex.

	if (elem->getFirstChild()->getFirstChild()) {
	  neighbour_[0] = elem->getFirstChild()->getFirstChild();
	  oppVertex_[0] = 2;

	  if (fill_opp_coords) {
            if (elem->getFirstChild()->isNewCoordSet()) {
	      oppCoord_[0] = *(elem->getFirstChild()->getNewCoord());
	    } else {
	      oppCoord_[0].setMidpoint(elInfoOld->coord_[0], 
				       elInfoOld->coord_[2]);
	    }

	    neighbourCoord_[0][0] = coord_[2];
	    neighbourCoord_[0][1] = coord_[1];
	    neighbourCoord_[0][2] = oppCoord_[0];
	  }
	} else {
	  neighbour_[0] = elem->getFirstChild();
	  oppVertex_[0] = 1;

	  if (fill_opp_coords) {
	    oppCoord_[0] = elInfoOld->coord_[0];

	    neighbourCoord_[0][0] = elInfoOld->coord_[2];
	    neighbourCoord_[0][1] = elInfoOld->coord_[0];
	    neighbourCoord_[0][2] = coord_[2];
	  }
	}

	// Calculation of the neighbour 1, its oppCoords and the
	// cooresponding oppVertex.

	nb = elInfoOld->neighbour_[2];
	if (nb) {
	  TEST(elInfoOld->oppVertex_[2] == 2)("invalid neighbour\n"); 
	  TEST((nb = nb->getFirstChild()))("missing child?\n");

	  if (nb->getFirstChild()) {
	    oppVertex_[1] = 2;

	    if (fill_opp_coords) {
	      if (nb->isNewCoordSet()) {
		oppCoord_[1] = *(nb->getNewCoord());
	      } else {
		oppCoord_[1].setMidpoint(elInfoOld->neighbourCoord_[2][0],
					 elInfoOld->neighbourCoord_[2][2]);
	      }

	      neighbourCoord_[1][0] = elInfoOld->neighbourCoord_[2][0];
	      neighbourCoord_[1][1].setMidpoint(elInfoOld->neighbourCoord_[2][0],
						elInfoOld->neighbourCoord_[2][1]);
	      neighbourCoord_[1][2] = oppCoord_[1];
	    }
	    nb = nb->getSecondChild();

	  } else {
	    oppVertex_[1] = 0;

	    if (fill_opp_coords) {
	      oppCoord_[1] = elInfoOld->oppCoord_[2];

	      neighbourCoord_[1][0] = elInfoOld->neighbourCoord_[2][2];	      
	      neighbourCoord_[1][1] = elInfoOld->neighbourCoord_[2][0];
	      neighbourCoord_[1][2].setMidpoint(elInfoOld->neighbourCoord_[2][0],
						elInfoOld->neighbourCoord_[2][1]);
	    }
	  }
	}
	neighbour_[1] = nb;
      } // if (ichild == 0) {} else
    } // if (fill_flag.isSet(Mesh::FILL_NEIGH) || fillFlag_.isSet(Mesh::FILL_OPP_COORDS))
    

    if (fill_flag.isSet(Mesh::FILL_BOUND)) {
      if (elInfoOld->getBoundary(2)) {
	boundary_[5] = elInfoOld->getBoundary(2);
      } else {
	boundary_[5] = INTERIOR;
      }

      if (ichild == 0) {
	boundary_[3] = elInfoOld->getBoundary(5);
	boundary_[4] = elInfoOld->getBoundary(3);
	boundary_[0] = elInfoOld->getBoundary(2);
	boundary_[1] = INTERIOR;
	boundary_[2] = elInfoOld->getBoundary(1);
      } else {
	boundary_[3] = elInfoOld->getBoundary(4);
	boundary_[4] = elInfoOld->getBoundary(5);
	boundary_[0] = INTERIOR;
	boundary_[1] = elInfoOld->getBoundary(2);
	boundary_[2] = elInfoOld->getBoundary(0);
      }

      if (elInfoOld->getProjection(0) && 
	  elInfoOld->getProjection(0)->getType() == VOLUME_PROJECTION) {
	
	projection_[0] = elInfoOld->getProjection(0);
      } else { // boundary projection
	if (ichild == 0) {
	  projection_[0] = elInfoOld->getProjection(2);
	  projection_[1] = NULL;
	  projection_[2] = elInfoOld->getProjection(1);
	} else {
	  projection_[0] = NULL;
	  projection_[1] = elInfoOld->getProjection(2);
	  projection_[2] = elInfoOld->getProjection(0);
	}
      }
    }
  }

  double ElInfo2d::calcGrdLambda(DimVec<WorldVector<double> >& grd_lam)
  {
    FUNCNAME("ElInfo2d::calcGrdLambda()");

    testFlag(Mesh::FILL_COORDS);
  
    double adet = 0.0;
    int dim = mesh_->getDim();

    for (int i = 0; i < dimOfWorld; i++) {
      (*e1)[i] = coord_[1][i] - coord_[0][i];
      (*e2)[i] = coord_[2][i] - coord_[0][i];
    }

    if (dimOfWorld == 2) {
      double sdet = (*e1)[0] * (*e2)[1] - (*e1)[1] * (*e2)[0];
      adet = abs(sdet);

      if (adet < 1.0E-25) {
	MSG("abs(det) = %f\n", adet);
	for (int i = 0; i <= dim; i++)
	  for (int j = 0; j < dimOfWorld; j++)
	    grd_lam[i][j] = 0.0;
      } else {
	double det1 = 1.0 / sdet;

	grd_lam[1][0] = (*e2)[1] * det1;  // a11: (a_ij) = A^{-T}
	grd_lam[1][1] = -(*e2)[0] * det1; // a21
	grd_lam[2][0] = -(*e1)[1] * det1; // a12
	grd_lam[2][1] = (*e1)[0] * det1;  // a22
	grd_lam[0][0] = - grd_lam[1][0] - grd_lam[2][0];
	grd_lam[0][1] = - grd_lam[1][1] - grd_lam[2][1];
      }
    } else {  
      vectorProduct(*e1, *e2, *normal);

      adet = norm(normal);

      if (adet < 1.0E-15) {
	MSG("abs(det) = %lf\n", adet);
	for (int i = 0; i <= dim; i++)
	  for (int j = 0; j < dimOfWorld; j++)
	    grd_lam[i][j] = 0.0;
      } else {
	vectorProduct(*e2, *normal, grd_lam[1]);
	vectorProduct(*normal, *e1, grd_lam[2]);
      
	double adet2 = 1.0 / (adet * adet);

	for (int i = 0; i < dimOfWorld; i++) {
	  grd_lam[1][i] *= adet2;
	  grd_lam[2][i] *= adet2;
	}

	grd_lam[0][0] = - grd_lam[1][0] - grd_lam[2][0];
	grd_lam[0][1] = - grd_lam[1][1] - grd_lam[2][1];
	grd_lam[0][2] = - grd_lam[1][2] - grd_lam[2][2];
      }
    }

    return adet;
  }

  const int ElInfo2d::worldToCoord(const WorldVector<double>& xy,
				   DimVec<double>* lambda) const
  {
    FUNCNAME("ElInfo::worldToCoord()");

    TEST_EXIT_DBG(lambda)("lambda must not be NULL\n");

    DimVec<WorldVector<double> > edge(mesh_->getDim(), NO_INIT);
    WorldVector<double> x; 
    static DimVec<double> vec(mesh_->getDim(), NO_INIT);

    int dim = mesh_->getDim();

    for (int j = 0; j < dimOfWorld; j++) {
      double x0 = coord_[dim][j];
      x[j] = xy[j] - x0;
      for (int i = 0; i < dim; i++)
	edge[i][j] = coord_[i][j] - x0;
    }
  
    double det  = edge[0][0] * edge[1][1] - edge[0][1] * edge[1][0]; 
    double det0 =       x[0] * edge[1][1] -       x[1] * edge[1][0]; 
    double det1 = edge[0][0] * x[1]       - edge[0][1] * x[0]; 

    if (abs(det) < DBL_TOL) {
      ERROR("det = %le; abort\n", det);
      for (int i = 0; i <= dim; i++) 
	(*lambda)[i] = 1.0/dim;
      return 0;
    }

    (*lambda)[0] = det0 / det;
    (*lambda)[1] = det1 / det;
    (*lambda)[2] = 1.0 - (*lambda)[0] - (*lambda)[1];

    int k = -1;
    double lmin = 0.0;
    for (int i = 0; i <= dim; i++) {
      if ((*lambda)[i] < -1.E-5) {
	if ((*lambda)[i] < lmin) {
	  k = i;
	  lmin = (*lambda)[i];
	}
      }
    }

    return k;
  }


  double ElInfo2d::getNormal(int side, WorldVector<double> &normal)
  {
    FUNCNAME("ElInfo::getNormal()");

    int i0 = (side + 1) % 3;
    int i1 = (side + 2) % 3;

    if (dimOfWorld == 2){
      normal[0] = coord_[i1][1] - coord_[i0][1];
      normal[1] = coord_[i0][0] - coord_[i1][0];
    } else { // dow == 3
      WorldVector<double> e0, e1,e2, elementNormal;

      e0 = coord_[i1]; 
      e0 -= coord_[i0];
      e1 = coord_[i1]; 
      e1 -= coord_[side];
      e2 = coord_[i0]; 
      e2 -= coord_[side];

      vectorProduct(e1, e2, elementNormal);
      vectorProduct(elementNormal, e0, normal);
    }

    double det = norm(&normal);

    TEST_EXIT_DBG(det > 1.e-30)("det = 0 on face %d\n", side);

    normal *= 1.0 / det;
    
    return det;
  }

  /****************************************************************************/
  /*  calculate the normal of the element for dim of world = dim + 1          */
  /*  return the absulute value of the determinant from the                   */
  /*  transformation to the reference element                                 */
  /****************************************************************************/
  double ElInfo2d::getElementNormal(WorldVector<double> &elementNormal) const
  {
    FUNCNAME("ElInfo::getElementNormal()");

    TEST_EXIT_DBG(dimOfWorld == 3)
      (" element normal only well defined for  DIM_OF_WORLD = DIM + 1 !!");

    WorldVector<double> e0 = coord_[1] - coord_[0];
    WorldVector<double> e1 = coord_[2] - coord_[0];

    vectorProduct(e0, e1, elementNormal);

    double det = norm(&elementNormal);

    TEST_EXIT_DBG(det > 1.e-30)("det = 0");

    elementNormal *= 1.0 / det;
    
    return det;
  }

}
