// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ElementDofIterator.h */

#ifndef AMDIS_ELEMENTDOFITERATOR_H
#define AMDIS_ELEMENTDOFITERATOR_H

#include "AMDiS_fwd.h"
#include "Global.h"
#include "Mesh.h"

namespace AMDiS {

  /** \brief
   * This class implements an iterator to iterate over all dofs of one element
   * independet of dimension and the degree of basis functions.
   *
   * Should be used in the following way:
   *
   *    ElementDofIterator elDofIter(admin);
   *    elDofIter.reset(el);
   *    do {
   *       elDofIter.getDOF();
   *    } while (elDofIter.next());
   *
   */
  class ElementDofIterator
  {
  public:
    /// Constructor.
    ElementDofIterator(const FiniteElemSpace* feSpace, bool inOrderPos = false)
      : admin(feSpace->getAdmin()),
	basisFcts(feSpace->getBasisFcts()),
	mesh(feSpace->getMesh()),
	dim(mesh->getDim()),
	inOrder(inOrderPos)
    {}

    /// Start a new traverse with the given element.
    void reset(Element* el);

    /// Go to next dof. Returns false, if there is dof anymore.
    bool next();

    bool nextStrict();

    /// Returns the dof index of the current dof.
    inline const DegreeOfFreedom getDof()
    {
      if (inOrder) 
	return dofs[node0 + elementPos][n0 + orderPosition[dofPos]];
      else
	return dofs[node0 + elementPos][n0 + dofPos];
    }
    
    /// Returns a pointer to the current dof.
    inline const DegreeOfFreedom* getDofPtr()
    {
      if (inOrder)
	return &dofs[node0 + elementPos][n0 + orderPosition[dofPos]];
      else
	return &dofs[node0 + elementPos][n0 + dofPos];
    }

    /// Returns \ref pos, the current position (vertex, edge, face) of the traverse.
    inline int getCurrentPos()
    {
      return pos;
    }

    /// Returns \ref elementPos, the number of vertex, edge or face that is traversed.
    inline int getCurrentElementPos()
    {
      return elementPos;
    }

    inline GeoIndex getPosIndex()
    {
      return posIndex;
    }

      
  protected:
    /// The dof admin for which dofs should be traversed.
    const DOFAdmin* admin;

    const BasisFunction* basisFcts;

    /// Pointer to the dofs that should be traversed.
    const DegreeOfFreedom **dofs;

    /// Mesh on which the element is defined.
    Mesh* mesh;

    /// Dimension of the mesh.
    int dim;

    bool inOrder;

    int* orderPosition;

    Element* element;

    /// Current position (i.e., vertex, edge, face) of the traverse.
    int pos;

    /// Dimension dependent geo index of the current position in traverse.
    GeoIndex posIndex;

    /** \brief
     * Number of dofs at the current traverse position. Examples: independent of 
     * dimension and  degree of basis functions there is only one dof per vertex. 
     * But in 2d and with 3rd degree lagrange basis functions there are two 
     * dofs per edge.
     */
    int nDofs;

    /** \brief
     * Displacement of dof indices. Used of more than one dof admin is defined 
     * on the mesh.
     */
    int n0;

    /// Dof index of the first dof at this geo index position.
    int node0;
    
    /** \brief
     * Number of elements in the current geo position. Examples: 3 vertices in 2d,
     * 1 face in 2d, 4 faces in 3d, etc.
     */
    int nElements;

    /// Current element, i.e., ith vertex, edge or face, that is traversed.
    int elementPos;

    /// Currrent dof that is traversed on the current element;
    int dofPos;
  };
}

#endif // AMDIS_ELEMENTDOFITERATOR_H
