#include "Mesh.h"
#include "Parametric.h"
#include "Quadrature.h"

namespace AMDiS {

  template<typename T>
  T Error<T>::errUFct(const DimVec<double>& lambda)
  {
    WorldVector<double> x;
    elinfo->coordToWorld(lambda, x);
    return ((*pU)(x));
  }

  template<typename T>
  WorldVector<T> Error<T>::grdErrUFct(const DimVec<double>& lambda)
  {
    WorldVector<double> x;
    elinfo->coordToWorld(lambda, x);
    return ((*pGrdU)(x));
  }


  template<typename T>
  int Error<T>::maxErrAtQpFct(ElInfo* el_info)
  {
    double err = 0.0;
    const double *u_vec, *uh_vec;

    elinfo = el_info;
    u_vec = quadFast->getQuadrature()->fAtQp(errU, NULL);
    uh_vec = errUh->getVecAtQPs(el_info,
				NULL,
				quadFast,
				NULL);

    int numPoints = quadFast->getNumPoints();

    for (int i = 0; i < numPoints; i++) {
      err = u_vec[i] > uh_vec[i] ? u_vec[i] - uh_vec[i] : uh_vec[i] - u_vec[i];
      maxErr = max(maxErr, err);
    }

    return;  
  }

  template<typename T>
  double Error<T>::maxErrAtQp(const AbstractFunction<T, WorldVector<double> >& u,
			      const DOFVector<T>& uh,
			      const Quadrature* q)
  {
    FUNCNAME("Error<T>::maxErrAtQp()");
    const FiniteElemSpace *fe_space;

    if (!(pU = &u)) {
      ERROR("no function u specified; doing nothing\n");
      return(-1.0);
    }

    if (!(errUh = &uh)  ||  !(fe_space = uh->getFESpace()))
      {
	ERROR("no discrete function or no fe_space for it; doing nothing\n");
	return(-1.0);
      }

    if (!(basFct = fe_space->getBasisFcts()))
      {
	ERROR("no basis functions at discrete solution ; doing nothing\n");
	return(-1.0);
      }

    int dim = fe_space->getMesh()->getDim();

    if (!q)
      q = Quadrature::provideQuadrature(dim, 
					2*fe_space->getBasisFcts()->getDegree()
					- 2);
    quadFast = FastQuadrature::provideFastQuadrature(basFct, *q, INIT_PHI);

    maxErr = 0.0;

    fe_space->getMesh()->traverse(-1, 
				  Mesh::FILL_COORDS |
				  Mesh::CALL_LEAF_EL, 
				  maxErrAtQpFct);

    return(maxErr);
  }

  template<typename T>
  int Error<T>::H1ErrFct(ElInfo* el_info)
  {
    int i, j;
    double err, err_2, h1_err_el, norm_el, norm2, det, exact;
    const WorldVector<double> *grdu_vec, *grduh_vec;

    elinfo = el_info;
    grdu_vec = quadFast->getQuadrature()->grdFAtQp(grdErrU, NULL);
    det = el_info->getDet();
    grduh_vec = errUh->getGrdAtQPs(elinfo, NULL, quadFast, NULL);

    int numPoints = quadFast->getNumPoints();
    int dow = Global::getGeo(WORLD);

    for (h1_err_el = i = 0; i < numPoints; i++)
      {
	for (err_2 = j = 0; j < dow; j++) 
	  {
	    err = grdu_vec[i][j] - grduh_vec[i][j];
	    err_2 += sqr(err);
	  }
	h1_err_el += quadFast->getWeight(i)*err_2;
      }

    exact = det*h1_err_el;
    h1Err2 += exact;
    maxErr = max(maxErr, exact);

    if (writeInLeafData)
      el_info->getElement()->setEstimation(exact, component);

    if (relative) {
      for (norm_el = i = 0; i < numPoints; i++) {
	for (norm2 = j = 0; j < dow; j++) 
	  norm2 += sqr(grdu_vec[i][j]);
	norm_el += quadFast->getWeight(i)*norm2;
      }
      h1Norm2 += det*norm_el;
    }

    return 0;
  }

  template<typename T>
  double Error<T>::H1Err(
			 const AbstractFunction<WorldVector<T>, WorldVector<double> >& grdU,
			 const DOFVector<T>&                                     uh,
			 int                                                     relErr,
			 double*                                                 max,
			 bool writeLeafData,
			 int comp)
  {
    FUNCNAME("Error<T>::H1Err");
    const FiniteElemSpace *fe_space;
  
    writeInLeafData = writeLeafData;

    component = comp;

    Quadrature *q = NULL;

    pGrdU = &grdU;

    errUh = &uh;

    if(!(fe_space = uh.getFESpace()))
      {
	ERROR("no fe_space for uh; doing nothing\n");
	return(0.0);
      }

    if (!(basFct = fe_space->getBasisFcts()))
      {
	ERROR("no basis functions at discrete solution ; doing nothing\n");
	return(0.0);
      }

    int dim = fe_space->getMesh()->getDim();
    int deg = grdU.getDegree();
    int degree = deg ? deg : 2 * fe_space->getBasisFcts()->getDegree() - 2;

    q = Quadrature::provideQuadrature(dim, degree);
    quadFast = FastQuadrature::provideFastQuadrature(basFct, 
						     *q, 
						     INIT_GRD_PHI);

    relative = relErr;

    maxErr = h1Err2 = h1Norm2 = 0.0;


    fe_space->getMesh()->traverse(-1, 
				  Mesh::FILL_COORDS |
				  Mesh::CALL_LEAF_EL |
				  Mesh::FILL_DET | 
				  Mesh::FILL_GRD_LAMBDA, 
				  H1ErrFct);

    if (relative) {
      relNorm2 = h1Norm2 + 1.e-15;

      fe_space->getMesh()->traverse(-1, Mesh::CALL_LEAF_EL, relFct);

      h1Err2 /= relNorm2;
      maxErr /= relNorm2;
    }

    if (max)  *max = maxErr;

    return(sqrt(h1Err2));
  }

  template<typename T>
  int Error<T>::L2ErrFct(ElInfo* el_info)
  {
    int i;
    double err, det, l2_err_el, norm_el, exact;
    const double *u_vec, *uh_vec;
    elinfo = el_info;

    u_vec = quadFast->getQuadrature()->fAtQp(errU, NULL);
    uh_vec = errUh->getVecAtQPs(el_info,
				NULL,
				quadFast,
				NULL);
    det = el_info->getDet();
  
    int numPoints = quadFast->getNumPoints();
  
    for (l2_err_el = i = 0; i < numPoints; i++) {
      err = u_vec[i] - uh_vec[i];
      l2_err_el += quadFast->getWeight(i)*sqr(err);
    }

    exact = det*l2_err_el;
    l2Err2 += exact;
    maxErr = max(maxErr, exact);

    if (writeInLeafData) {
      el_info->getElement()->setEstimation(exact, component);
    }

    if (relative) {
      for (norm_el = i = 0; i < numPoints; i++)
	norm_el += quadFast->getWeight(i)*sqr(u_vec[i]);
      l2Norm2 += det*norm_el;
    }

    return 0;
  }

  template<typename T>
  double Error<T>::L2Err(const AbstractFunction<T, WorldVector<double> >& u,
			 const DOFVector<T>& uh,
			 int relErr,
			 double* max,
			 bool writeLeafData,
			 int comp)
  {
    FUNCNAME("Error<T>::L2Err");
    const FiniteElemSpace *fe_space;

    Quadrature *q = NULL;

    writeInLeafData = writeLeafData;
  
    component = comp;

    if (!(pU = &u)) {
      ERROR("no function u specified; doing nothing\n");
      return(0.0);
    }
    
    if (!(errUh = &uh)  ||  !(fe_space = uh.getFESpace())) {
      ERROR("no discrete function or no fe_space for it; doing nothing\n");
      return(0.0);
    }

    if (!(basFct = fe_space->getBasisFcts())) {
      ERROR("no basis functions at discrete solution ; doing nothing\n");
      return(0.0);
    }

    int dim = fe_space->getMesh()->getDim();
    int deg = u.getDegree();
    int degree = deg ? deg :  2 * fe_space->getBasisFcts()->getDegree() - 2;

    q = Quadrature::provideQuadrature(dim, degree);
    quadFast = FastQuadrature::provideFastQuadrature(basFct, *q, INIT_PHI);

    relative = relErr;

    maxErr = l2Err2 = l2Norm2 = 0.0;

    fe_space->getMesh()->traverse(-1, 
				  Mesh::FILL_COORDS |
				  Mesh::FILL_DET |
				  Mesh::FILL_GRD_LAMBDA |
				  Mesh::CALL_LEAF_EL,
				  L2ErrFct);

    if (relative) {
      relNorm2 = l2Norm2 + 1.e-15;
      fe_space->getMesh()->traverse(-1, Mesh::CALL_LEAF_EL, relFct);
      l2Err2 /= relNorm2;
    }

    if (max)  
      *max = maxErr;

    return (sqrt(l2Err2));
  }

  template<typename T>
  int Error<T>::relFct(ElInfo* elinfo)
  {
    double exact = elinfo->getElement()->getEstimation(component);
    exact /= relNorm2;
    if (writeInLeafData)
      elinfo->getElement()->setEstimation(exact, component);
    return 0;
  }

}
