#include "MacroElement.h"
#include "Boundary.h"
#include "FiniteElemSpace.h"
#include "Mesh.h"
#include <string>
#include "FixVec.h"
#include "FixVecConvert.h"
#include <map>

namespace AMDiS {

  MacroElement::MacroElement(int dim)
    : element(NULL),
      coord(dim, NO_INIT),
      boundary(dim, DEFAULT_VALUE, INTERIOR),
      projection(dim, NO_INIT),
      neighbour(dim, NO_INIT),
      oppVertex(dim, NO_INIT),
      index(-1), 
      elType(0),
      deserializedNeighbourIndices_(NULL)
  {
    neighbour.set(NULL);
    projection.set(NULL);
  }

  MacroElement::~MacroElement()
  {
    if (element)
      delete element;
  }

  MacroElement& MacroElement::operator=(const MacroElement &el)
  {
    FUNCNAME("MacroElement::operator=()");

    if (this == &el)
      return *this;

    coord = el.coord;
    boundary = el.boundary;
    projection = el.projection;
    oppVertex = el.oppVertex;
    index = el.index;
    elType = el.elType;  
    
    return *this;
  }

  void MacroElement::serialize(std::ostream &out)
  {
    // write element-tree
    out << element->getTypeName() << "\n";
    element->serialize(out);

    // write coords
    int size = coord.getSize();
    out.write(reinterpret_cast<const char*>(&size), sizeof(int));
    for (int i = 0; i < size; i++)
      coord[i].serialize(out);

    // write boundary
    boundary.serialize(out);

    // write projection
    size = projection.getSize();
    out.write(reinterpret_cast<const char*>(&size), sizeof(int));
    for (int i = 0; i < size; i++) {
      int id = projection[i] ? projection[i]->getID() : -1;
      out.write(reinterpret_cast<const char*>(&id), sizeof(int));
    }

    // write neighbour
    size = neighbour.getSize();
    out.write(reinterpret_cast<const char*>(&size), sizeof(int));
    for (int i = 0; i < size; i++) {
      int index = neighbour[i] ? neighbour[i]->getIndex() : -1;
      out.write(reinterpret_cast<const char*>(&index), sizeof(int));
    }
  
    // write oppVertex
    oppVertex.serialize(out);

    // write index
    out.write(reinterpret_cast<const char*>(&index), sizeof(int));

    // write elType
    out.write(reinterpret_cast<const char*>(&elType), sizeof(unsigned char));
  }

  void MacroElement::deserialize(std::istream &in)
  {
    // read element-tree
    std::string typeName;
    in >> typeName;
    in.get();

    if (element) {
      TEST_EXIT(typeName == element->getTypeName())("wrong element type name\n");
    } else {
      if (typeName == "Line") 
	element = new Line(NULL);
      if (typeName == "Triangle") 
	element = new Triangle(NULL);
      if (typeName == "Tetrahedron") 
	element = new Tetrahedron(NULL);
    }

    element->deserialize(in);

    // read coords
    int size;
    in.read(reinterpret_cast<char*>(&size), sizeof(int));
    if (coord.getSize()) {
      TEST_EXIT(coord.getSize() == size)("invalid size\n");
    } else {
      coord.initSize(size);
    }
    for (int i = 0; i < size; i++) {
      coord[i].deserialize(in);
    }

    // read boundary
    boundary.deserialize(in);

    // read projection
    in.read(reinterpret_cast<char*>(&size), sizeof(int));

    if (projection.getSize()) {
      TEST_EXIT(projection.getSize() == size)("invalid size\n");
    } else {
      projection.initSize(size);
    }

    for (int i = 0; i < size; i++) {
      int id;
      in.read(reinterpret_cast<char*>(&id), sizeof(int));
      projection[i] = (id != -1) ? Projection::getProjection(id) : NULL;
    }

    // read neighbour indices
    in.read(reinterpret_cast<char*>(&size), sizeof(int));
  
    TEST_EXIT(deserializedNeighbourIndices_)
      ("neighbour indices for deserializing not set\n");
    
    deserializedNeighbourIndices_->resize(size);

    if(neighbour.getSize()) {
      TEST_EXIT(neighbour.getSize() == size)("invalid size\n");
    } else {
      neighbour.initSize(size);
    }

    for (int i = 0; i < size; i++) {
      int index;
      in.read(reinterpret_cast<char*>(&index), sizeof(int));

      //     neighbour[i] = reinterpret_cast<MacroElement*>(index);
      //     // attention: the neighbour INDEX is stored in the neighbour vector.
      //     // After all macro elements are read in, the indices must be replaced
      //     // by the corresponding pointers.

      (*deserializedNeighbourIndices_)[i] = index;
    }

    deserializedNeighbourIndices_ = NULL;

    // read oppVertex
    oppVertex.deserialize(in);

    // write index
    in.read(reinterpret_cast<char*>(&index), sizeof(int));


    // write elType
    in.read(reinterpret_cast<char*>(&elType), sizeof(unsigned char));
  }

  int MacroElement::calcMemoryUsage() 
  {
    int result = 0;

    result += sizeof(MacroElement);
    result += coord.getSize() * sizeof(WorldVector<double>);
    result += projection.getSize() * sizeof(Projection*);
    result += neighbour.getSize() * sizeof(MacroElement*);
    result += oppVertex.getSize() * sizeof(char);
    result += element->calcMemoryUsage();

    return result;
  }

}
