// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ParMetisPartitioner.h */

#ifndef AMDIS_PARMETISADAPTER_H
#define AMDIS_PARMETISADAPTER_H

#include <map>
#include <set>
#include "parmetis.h"
#include <Global.h>
namespace AMDiS {

  class FiniteElemSpace;

  enum PartitionMode {
    INITIAL = 0,          // initial partitioning of a unpartitioned mesh 
    ADAPTIVE_REPART = 1,  // repartitioning of a adaptively refined mesh
    REFINE_PART = 2       // quality improvement of the current partitioning
  };

  class ParMetisGraph;
  class Mesh;
  template<typename T > class DOFVector;

  class ParMetisMesh
  {
  public:
    ParMetisMesh(Mesh *mesh, MPI::Comm *comm);

    ~ParMetisMesh();

    inline void setParMetisIndex(int amdisIndex, int parMetisIndex) {
      elem_a2p[amdisIndex] = parMetisIndex + 1;
    }

    inline int getParMetisIndex(int amdisIndex) {
      int result = elem_a2p[amdisIndex];
      TEST_EXIT(result > 0)("invalid index\n");
      return result - 1;
    }

    inline void setAMDiSIndex(int parMetisIndex, int amdisIndex) {
      elem_p2a[parMetisIndex] = amdisIndex;
    }

    inline int getAMDiSIndex(int parMetisIndex) {
      return elem_p2a[parMetisIndex];
    }

    inline int *getAMDiSIndices() {
      return elem_p2a;
    }

    inline int *getElementPtr() { 
      return eptr; 
    }

    inline int *getElementInd() { 
      return eind; 
    }

    inline int *getElementDist() { 
      return elmdist; 
    }

    inline int getDim() { 
      return dim; 
    }

    inline float *getXYZ() { 
      return xyz; 
    }

    inline int getNumElements() { 
      return nElements; 
    }

  protected:
    int *eptr;

    int *eind;

    /* \brief
     * Array that specifies the distribution of the mesh elements.
     *
     * elmdist[0] = 0;
     * elmdist[1] = number of elements of proc 0;
     * elmdist[2] = elmdist[1] + number of elements of proc 1;
     *    ...
     */
    int *elmdist;

    int dim;

    float *xyz;

    int nElements;

    std::map<int, int> elem_a2p;

    int *elem_p2a;

    /// The MPI communicator that should be used for mesh partition.
    MPI::Comm *mpiComm;
  };


  class ParMetisGraph
  {
  public:
    ParMetisGraph(ParMetisMesh *parMetisMesh,
		  MPI::Comm *comm,
		  int ncommonnodes = -1);

    ~ParMetisGraph();

    inline int *getXAdj() { 
      return xadj; 
    }

    inline int *getAdjncy() { 
      return adjncy; 
    }

  protected:
    ParMetisMesh *parMetisMesh;

    int *xadj;

    int *adjncy;
  };


  class ParMetisPartitioner
  {
  public:
    ParMetisPartitioner(Mesh *mesh, MPI::Comm *comm)
      : mesh_(mesh),
        mpiComm(comm),
	parMetisMesh(NULL)
    {}

    void partition(std::map<int, double> *elemWeights,
		   PartitionMode mode = INITIAL,
		   float itr = 1000000.0);

    /* \brief 
     * Creates a map which stores for each element on partitioning level the 
     * rank the element corresponds to.
     */
    void fillCoarsePartitionVec(std::map<int, int> *partitionVec);

    void fillLeafPartitionVec(std::map<int, int> *coarseVec,
			      std::map<int, int> *fineVec);

    /* \brief
     * Creates an initial paritioning of the AMDiS mesh by seting the partition status
     * of all elements to either IN or UNDEFINED.
     */
    void createPartitionData();

    void deletePartitionData();

  protected:
    void distributePartitioning(int *part);

    void descendPartitionData(Element *element);

  protected:
    Mesh *mesh_;

    MPI::Comm *mpiComm;

    ParMetisMesh *parMetisMesh;
  };
}

#endif
