// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file QPsiPhi.h */

#ifndef AMDIS_ASSEMBLE_H
#define AMDIS_ASSEMBLE_H

#include <list>
#include "AMDiS_fwd.h"
#include "DOFAdmin.h"

namespace AMDiS {

  /**
   * \ingroup Integration
   *
   * \brief
   * Used for the comparison of two QPsiPhi objects in find_if.
   */
  template<typename T>
  class compareQPsiPhi : public std::unary_function<bool, T*>
  {
  public:
    /** \brief
     * Constructor.
     */
    compareQPsiPhi(const BasisFunction* psi_, 
		   const BasisFunction* phi_, 
		   const Quadrature* quad_) 
      : psi(psi_), phi(phi_), quadrature(quad_)
    {}

    /** \brief
     * Returns true, if *q is equivalent to *this.
     */
    bool operator()(T* q) {
      return (q->psi == psi) && (q->phi == phi) && (q->quadrature == quadrature);
    }
  
  private:
    /** \brief
     * Basis functions of QPsiPhi.
     */
    const BasisFunction *psi, *phi;

    /** \brief
     * Quadrature of QPsiPhi.
     */
    const Quadrature *quadrature;
  };

  // ============================================================================

  /**
   * \ingroup Integration
   *
   * \brief
   * Used for the comparison of two QPsi objects in find_if.
   */
  template<typename T>
  class compareQPsi : public std::unary_function<bool, T*>
  {
  public:
    /** \brief
     * Constructor.
     */
    compareQPsi(const BasisFunction* psi_, 
		const Quadrature* quad_) 
      : psi(psi_), quadrature(quad_)
    {}

    /** \brief
     * Returns true, if *q is equivalent to *this.
     */
    bool operator()(T* q) {
      return (q->psi == psi) && (q->quadrature == quadrature);
    }
  
  private:
    /** \brief
     * Basis functions of the QPsi object.
     */
    const BasisFunction *psi;

    /** \brief
     * Quadrature of the QPsi object.
     */
    const Quadrature *quadrature;
  };

  // ===========================================================================
  // ===== class Q11PsiPhi =====================================================
  // ===========================================================================

  /** \ingroup Integration
   * \brief
   * Calculates element stiffness matrices by preevaluated integrals over the 
   * the reference element (integral of the product of the derivatives of psi 
   * and phi).
   */
  class Q11PsiPhi
  {
  protected:
    /** \brief
     * Constructor
     */
    Q11PsiPhi(const BasisFunction *psi,
	      const BasisFunction *phi, 
	      const Quadrature    *q);


  public:
    /** \brief
     * Destructor
     */
    ~Q11PsiPhi();

    /** \brief
     * Returns a Q11PsiPhi object.
     */
    static const Q11PsiPhi* provideQ11PsiPhi(const BasisFunction *,
					     const BasisFunction *, 
					     const Quadrature*);
    
    /** \brief
     * Compares two Q11PsiPhi objects.
     */
    const bool operator==(const Q11PsiPhi&) const;

    /** \brief
     * Compares two Q11PsiPhi objects.
     */
    const bool operator!=(const Q11PsiPhi& q11pp) const { 
      return !(operator==(q11pp));
    }
    
    /** \brief
     * Returns \ref values[i][j][k]
     */
    inline const double getValue(unsigned int i,
				 unsigned int j,
				 unsigned int v) const 
    {
      if ((values)&&(values[i])&&(values[i][j])&&((static_cast<int>(v))<nrEntries[i][j]))
	return values[i][j][v];
      return 0.;
    }

    /** \brief
     * Returns \ref nrEntries[i][j]
     */
    inline const int getNumberEntries(unsigned int i, unsigned int j) const
    {
      if ((nrEntries)&&(nrEntries[i])) return nrEntries[i][j];
      return 0;
    }

    /** \brief
     * Returns \ref nrEntries
     */
    inline const int** getNumberEntries() const {
      return const_cast<const int**>( nrEntries);
    }

    /** \brief
     * Returns \ref k[i1][i2][i3]
     */
    inline const int 
    getK(unsigned int i1, unsigned int i2, unsigned int i3) const
    {
      if ((k)&&(k[i1])&&(k[i1][i2])&&((static_cast<int>(i3))<nrEntries[i1][i2])) 
	return k[i1][i2][i3];
      return 0;
    };

    /** \brief
     * Returns \ref l[i][j][v]
     */
    inline const int getL(unsigned int i, unsigned int j, unsigned int v) const
    {
      if ((l)&&(l[i])&&(l[i][j])&&((static_cast<int>(v))<nrEntries[i][j])) return l[i][j][v];
      return 0;
    }

    /** \brief
     * Returns \values[i][j]
     */
    inline const double *getValVec(unsigned int i, unsigned int j) const {
      if ((values)&&(values[i])&&(values[i][j])) return values[i][j];
      return NULL;
    }

    /** \brief
     * Returns \ref k[i][j]
     */
    inline const int *getKVec(unsigned int i, unsigned int j) const {
      if ((k)&&(k[i])&&(k[i][j])) return k[i][j];
      return NULL;
    }

    /** \brief
     * Returns \ref l[i][j] 
     */
    inline const int *getLVec(unsigned int i,unsigned int j) const {
      if ((l)&&(l[i])&&(l[i][j])) return l[i][j];
      return NULL;
    }

  protected:
    /** \brief
     * List of pointers to all Q11PsiPhi objects
     */
    static std::list<Q11PsiPhi*> preList;

    /** \brief
     * Pointer to the first set of basis functions
     */
    const BasisFunction *psi;

    /** \brief
     * pointer to the second set of basis functions
     */
    const BasisFunction *phi;

    /** \brief
     * Pointer to the Quadrature which is used for the integration
     */
    const Quadrature *quadrature;
  
    /** \brief
     * Matrix of size psi->getNumber() * phi->getNumber() storing the count of 
     * non zero integrals; nrEntries[i][j] is the count of non zero values of 
     * \f$ \hat{Q}_{ij,kl}^{11} \f$
     * (0 <= k,l <= DIM) for the pair (psi[i], phi[j]), 0 <= i < 
     * psi->getNumber(), 0 <= j < phi->getNumber()
     */
    int **nrEntries;

    /** \brief
     * tensor storing the non zero integrals; values[i][j] is a vector of length
     * \ref nrEntries[i][j] storing the non zero values for the pair (psi[i], 
     * phi[j])
     */
    double ***values;

    /** \brief
     * tensor storing the indices k of the non zero integrals;
     */
    int ***k;

    /** \brief
     * tensor storing the indices l of the non zero integrals;
     */
    int ***l;

    /** \brief
     * Number of non zero entries.
     */
    int allEntries;

    /** \brief
     * Pointer to an array in values.
     */
    double *val_vec;

    /** \brief
     * Pointer to an array in k.
     */
    int *k_vec;

    /** \brief
     * Pointer to an array in l.
     */
    int *l_vec;

    friend class compareQPsiPhi<Q11PsiPhi>;
  };



  // ===========================================================================
  // ===== class Q10PsiPhi =====================================================
  // ===========================================================================

  /** \ingroup Integration 
   * \brief
   * Calculates element stiffness matrices by preevaluated integrals over the 
   * the reference element (integral of the product of the derivative of psi 
   * and phi).
   */
  class Q10PsiPhi
  {
  protected:
    /** \brief
     * Constructor
     */
    Q10PsiPhi(const BasisFunction *psi,
	      const BasisFunction *phi, 
	      const Quadrature    *q);

  public:
    /** \brief
     * Destructor
     */
    ~Q10PsiPhi();

    /** \brief
     * Returns a Q10PsiPhi object.
     */
    static const Q10PsiPhi* provideQ10PsiPhi(const BasisFunction *,
					     const BasisFunction *, 
					     const Quadrature*);
    
    /** \brief
     * Compares two Q10PsiPhi objects.
     */
    const bool operator==(const Q10PsiPhi&) const;

    /** \brief
     * Compares two Q10PsiPhi objects.
     */
    const bool operator!=(const Q10PsiPhi& q10pp) const { 
      return !(operator==(q10pp));
    }

  
    /** \brief
     * Returns \ref values[i][j][k]
     */
    inline const double getValue(unsigned int i,
				 unsigned int j,
				 unsigned int v) const 
    {
      if ((values)&&(values[i])&&(values[i][j])&&((static_cast<int>(v))<nrEntries[i][j])) return values[i][j][v];
      return 0.;
    }

    /** \brief
     * Returns \ref nrEntries[i][j]
     */
    inline const int getNumberEntries(unsigned int i, unsigned int j) const {
      if ((nrEntries)&&(nrEntries[i])) return nrEntries[i][j];
      return 0;
    }

    /** \brief
     * Returns \ref nrEntries
     */
    inline const int** getNumberEntries() const {
      return const_cast<const int**>(nrEntries);
    }

    /** \brief
     * Returns \ref k[i1][i2][i3]
     */
    inline const int 
    getK(unsigned int i1, unsigned int i2, unsigned int i3) const 
    {
      if ((k)&&(k[i1])&&(k[i1][i2])&&((static_cast<int>(i3))<nrEntries[i1][i2])) 
	return k[i1][i2][i3];
      return 0;
    }

    /** \brief
     * Returns \values[i][j]
     */
    inline const double *getValVec(unsigned int i, unsigned int j) const {
      if ((values)&&(values[i])&&(values[i][j])) return values[i][j];
      return NULL;
    }

    /** \brief
     * Returns \ref k[i][j]
     */
    inline const int *getKVec(unsigned int i, unsigned int j) const {
      if ((k)&&(k[i])&&(k[i][j])) return k[i][j];
      return NULL;
    }

  protected:
    /** \brief
     * List of pointers to all Q11PsiPhi objects
     */
    static std::list<Q10PsiPhi*> preList;

    /** \brief
     * Pointer to the first set of basis functions
     */
    const BasisFunction *psi;

    /** \brief
     * pointer to the second set of basis functions
     */
    const BasisFunction *phi;

    /** \brief
     * Pointer to the Quadrature which is used for the integration
     */
    const Quadrature *quadrature;
  
    /** \brief
     * Matrix of size psi->getNumber() * phi->getNumber() storing the count of 
     * non zero integrals; nrEntries[i][j] is the count of non zero values of 
     * \f$ \hat{Q}_{ij,kl}^{11} \f$
     * (0 <= k,l <= DIM) for the pair (psi[i], phi[j]), 0 <= i < 
     * psi->getNumber(), 0 <= j < phi->getNumber()
     */
    int **nrEntries;

    /** \brief
     * tensor storing the non zero integrals; values[i][j] is a vector of length
     * \ref nrEntries[i][j] storing the non zero values for the pair (psi[i], 
     * phi[j])
     */
    double ***values;

    /** \brief
     * tensor storing the indices k of the non zero integrals;
     */
    int ***k;

    /** \brief
     * Number of all non zero entries.
     */
    int allEntries;

    /** \brief
     * Pointer to an array in values.
     */
    double *val_vec;

    /** \brief
     * Pointer to an array in k.
     */
    int *k_vec;

    friend class compareQPsiPhi<Q10PsiPhi>;
  };

  // ===========================================================================
  // ===== class Q01PsiPhi =====================================================
  // ===========================================================================

  /** \ingroup Integration 
   * \brief
   * Calculates element stiffness matrices by preevaluated integrals over the 
   * the reference element (integral of the product of psi and the derivative
   * of phi).
   */
  class Q01PsiPhi
  {
  protected:
    /** \brief
     * Constructor
     */
    Q01PsiPhi(const BasisFunction *psi,
	      const BasisFunction *phi, 
	      const Quadrature    *q);

  public:
    /** \brief
     * Destructor
     */
    ~Q01PsiPhi();

    /** \brief
     * Returns a Q01PsiPhi object.
     */
    static const Q01PsiPhi* provideQ01PsiPhi(const BasisFunction *,
					     const BasisFunction *, 
					     const Quadrature*);
    
    /** \brief
     * Compares two Q01PsiPhi objects.
     */
    const bool operator==(const Q01PsiPhi&) const;

    /** \brief
     * Compares two Q01PsiPhi objects.
     */
    const bool operator!=(const Q01PsiPhi& q01pp) const { 
      return !(operator==(q01pp));
    }

  
    /** \brief
     * Returns \ref values[i][j][k]
     */
    inline const double getValue(unsigned int i,
				 unsigned int j,
				 unsigned int v) const
    {
      if ((values)&&(values[i])&&(values[i][j])&&((static_cast<int>(v))<nrEntries[i][j])) 
	return values[i][j][v];
      return 0.;
    }

    /** \brief
     * Returns \ref nrEntries[i][j]
     */
    inline const int getNumberEntries(unsigned int i, unsigned int j) const
    {
      if ((nrEntries)&&(nrEntries[i])) return nrEntries[i][j];
      return 0;
    }

    /** \brief
     * Returns \ref nrEntries
     */
    inline const int** getNumberEntries() const {
      return const_cast<const int**>(nrEntries);
    }

    /** \brief
     * Returns \ref k[i1][i2][i3]
     */
    inline const int 
    getK(unsigned int i1, unsigned int i2, unsigned int i3) const;

    /** \brief
     * Returns \values[i][j]
     */
    inline const double *getValVec(unsigned int i, unsigned int j) const {
      if ((values)&&(values[i])&&(values[i][j])) return values[i][j];
      return NULL;
    }

    /** \brief
     * Returns \ref k[i][j]
     */
    inline const int *getLVec(unsigned int i, unsigned int j) const {
      if ((l)&&(l[i])&&(l[i][j])) return l[i][j];
      return NULL;
    }

    /** \brief
     * Returns \ref k[i][j][v]
     */
    inline const int getL(unsigned int i, 
			  unsigned int j, 
			  unsigned int v) const
    {
      if ((l)&&(l[i])&&(l[i][j])&&((static_cast<int>(v))<nrEntries[i][j])) return l[i][j][v];
      return 0;
    }

  protected:
    /** \brief
     * List of pointers to all Q11PsiPhi objects
     */
    static std::list<Q01PsiPhi*> preList;

    /** \brief
     * Pointer to the first set of basis functions
     */
    const BasisFunction *psi;

    /** \brief
     * pointer to the second set of basis functions
     */
    const BasisFunction *phi;

    /** \brief
     * Pointer to the Quadrature which is used for the integration
     */
    const Quadrature *quadrature;
  
    /** \brief
     * Matrix of size psi->getNumber() * phi->getNumber() storing the count of 
     * non zero integrals; nrEntries[i][j] is the count of non zero values of 
     * \f$ \hat{Q}_{ij,kl}^{11} \f$
     * (0 <= k,l <= DIM) for the pair (psi[i], phi[j]), 0 <= i < 
     * psi->getNumber(), 0 <= j < phi->getNumber()
     */
    int **nrEntries;

    /** \brief
     * tensor storing the non zero integrals; values[i][j] is a vector of length
     * \ref nrEntries[i][j] storing the non zero values for the pair (psi[i], 
     * phi[j])
     */
    double ***values;

    /** \brief
     * tensor storing the indices l of the non zero integrals;
     */
    int ***l;

    /** \brief
     * Number of all non zero entries.
     */
    int allEntries;

    /** \brief
     * Pointer to an array in values
     */
    double *val_vec;

    /** \brief
     * Pointer to an array in l.
     */
    int *l_vec;

    friend class compareQPsiPhi<Q01PsiPhi>;
  };

  // ===========================================================================
  // ===== class Q00PsiPhi =====================================================
  // ===========================================================================

  /** \ingroup Integration
   * \brief
   * Calculates element stiffness matrices by preevaluated integrals over the 
   * the reference element (integral of the product of psi and phi).
   */
  class Q00PsiPhi
  {
    /** \brief
     * List of pointers to all Q11PsiPhi objects
     */
    static std::list<Q00PsiPhi*> preList;

    /** \brief
     * Pointer to the first set of basis functions
     */
    const BasisFunction *psi;

    /** \brief
     * pointer to the second set of basis functions
     */
    const BasisFunction *phi;

    /** \brief
     * Pointer to the Quadrature which is used for the integration
     */
    const Quadrature *quadrature;

    /** \brief
     * Matrix storing the integrals
     * \f[ values[i][j] = \hat{Q}_{ij}^{00} = \hat{Q}(\overline{\psi}^i
     * \overline{\phi}^j) \f]
     * for the pair (psi[i], phi[j]), 0 <= i <= psi->getNumber(),
     * 0 <= j <= phi->getNumber()
     */  
    double  **values;
  
  protected:
    /** \brief
     * Constructor
     */
    Q00PsiPhi(const BasisFunction *,
	      const BasisFunction *, 
	      const Quadrature*);

  public:
    /** \brief
     * Destructor
     */
    ~Q00PsiPhi();

    /** \brief
     * Returns a Q00PsiPhi object.
     */
    static Q00PsiPhi* provideQ00PsiPhi(const BasisFunction *,
				       const BasisFunction *, 
				       const Quadrature*);
  
  
    /** \brief
     * Compares two Q00PsiPhi objects.
     */
    const bool operator==(const Q00PsiPhi&) const;

    /** \brief
     * Compares two Q00PsiPhi objects.
     */
    const bool operator!=(const Q00PsiPhi& q00pp) const { 
      return !(operator==(q00pp));
    }
  
    /** \brief
     * Returns \ref values[i][j]
     */
    const double getValue(unsigned int i,unsigned int j) const;

    /** \brief
     * Returns \ref values[i]
     */
    const double *getValVec(unsigned int i) const;  

    friend class compareQPsiPhi<Q00PsiPhi>;
  };


  // ===========================================================================
  // ===== class Q0Psi =========================================================
  // ===========================================================================

  /** \ingroup Integration
   * \brief
   * Integral of psi.
   */
  class Q0Psi
  {
  protected:
    /** \brief
     * List of pointers to all Q0Psi objects
     */
    static std::list<Q0Psi*> preList;

    /** \brief
     * Pointer to the first set of basis functions
     */
    const BasisFunction *psi;

    /** \brief
     * Pointer to the Quadrature which is used for the integration
     */
    const Quadrature *quadrature;

    /** \brief
     * Vector storing the integrals
     */
    double* values;
  
  protected:
    /** \brief
     * Constructor
     */
    Q0Psi(const BasisFunction *,
	  const Quadrature*);

  public:
    /** \brief
     * Destructor
     */
    ~Q0Psi();

    /** \brief
     * Returns a Q0Psi object.
     */
    static Q0Psi* provideQ0Psi(const BasisFunction *, const Quadrature*);
  
    /** \brief
     * Compares two Q0Psi objects.
     */
    const bool operator==(const Q0Psi&) const;

    /** \brief
     * Compares two Q0Psi objects.
     */
    const bool operator!=(const Q0Psi& q0p) const { 
      return !(operator==(q0p));
    }
  
    /** \brief
     * Returns \ref value
     */
    inline const double getValue(int i) const { 
      return values[i]; 
    }

    /** \brief
     * Returns \ref values[i]
     */
    inline const double *getValVec() const {
      return values;
    } 

    friend class compareQPsi<Q0Psi>;
  };

  // ===========================================================================
  // ===== class Q1Psi =========================================================
  // ===========================================================================

  /** \ingroup Integration
   * \brief
   * Integral of the derivative of psi.
   */
  class Q1Psi
  {
  protected:
    /** \brief
     * Constructor
     */
    Q1Psi(const BasisFunction *psi,
	  const Quadrature    *q);


  public:
    /** \brief
     * Destructor
     */
    ~Q1Psi();

    /** \brief
     * Returns a Q1Psi object.
     */
    static const Q1Psi* provideQ1Psi(const BasisFunction *,
				     const Quadrature*);
    
    /** \brief
     * Compares two Q1Psi objects.
     */
    const bool operator==(const Q1Psi&) const;

    /** \brief
     * Compares two Q1Psi objects.
     */
    const bool operator!=(const Q1Psi& q1p) const { 
      return !(operator==(q1p));
    }
    
    /** \brief
     * Returns \ref values[i][j]
     */
    inline const double getValue(unsigned int i,
				 unsigned int j) const 
    {
      if ((values)&&(values[i])&&((static_cast<int>(j))<nrEntries[i]))
	return values[i][j];
      return 0.;
    }

    /** \brief
     * Returns \ref nrEntries[i]
     */
    inline const int getNumberEntries(unsigned int i) const
    {
      if (nrEntries) return nrEntries[i];
      return 0;
    }

    /** \brief
     * Returns \ref nrEntries
     */
    inline const int* getNumberEntries() const {
      return const_cast<const int*>(nrEntries);
    }

    /** \brief
     * Returns \ref k[i1][i2]
     */
    inline const int 
    getK(unsigned int i1, unsigned int i2) const
    {
      if ((k)&&(k[i1])&&((static_cast<int>(i2))<nrEntries[i1])) 
	return k[i1][i2];
      return 0;
    };

    /** \brief
     * Returns \ref k[i]
     */
    inline const int *getKVec(unsigned int i) const {
      if ((k)&&(k[i])) return k[i];
      return NULL;
    }

    /** \brief
     * Returns \values[i]
     */
    inline const double *getValVec(unsigned int i) const {
      if ((values)&&(values[i])) return values[i];
      return NULL;
    }

  protected:
    /** \brief
     * List of pointers to all Q1Psi objects
     */
    static std::list<Q1Psi*> preList;

    /** \brief
     * Pointer to the first set of basis functions
     */
    const BasisFunction *psi;

    /** \brief
     * Pointer to the Quadrature which is used for the integration
     */
    const Quadrature *quadrature;
  
    /** \brief
     * Array of size psi->getNumber() storing the count of 
     * non zero integrals; nrEntries[i] is the count of non zero values of 
     * \f$ \hat{Q}_{i,kl}^{11} \f$
     * (0 <= k <= DIM) for psi[i], 0 <= i < psi->getNumber().
     */
    int *nrEntries;

    /** \brief
     * Number of all non zero entries.
     */
    int allEntries;

    /** \brief
     * tensor storing the non zero integrals; values[i] is a vector of length
     * \ref nrEntries[i] storing the non zero values for psi[i].
     */
    double **values;

    /** \brief
     * Matrix storing the indices k of the non zero integrals;
     */
    int **k;

    /** \brief
     * Pointer to an array in values.
     */
    double *val_vec;

    /** \brief
     * Pointer to an array in k.
     */
    int *k_vec;

    friend class compareQPsi<Q1Psi>;
  };

}

#endif // AMDIS_ASSEMBLE_H
