#include <iostream>
#include <sstream>

namespace AMDiS {

  template<typename T>
  SolutionDataStorage<T>::SolutionDataStorage(std::string name)
    : maxMemoryUsage(100),
      useDisk(true),
      writeDirectory(""),
      lastPos(-1),
      poped(false),
      memoryUsage(0),
      idcounter(0)
  {
    solutions.clear();
    
    int tmp = 0;
    GET_PARAMETER(0, name + "->memory usage", "%d", &maxMemoryUsage);
    GET_PARAMETER(0, name + "->use disk", "%d", &tmp);
    useDisk = (tmp == 0) ? false : true;
    GET_PARAMETER(0, name + "->directory", &writeDirectory);
  }

  template<typename T>
  SolutionDataStorage<T>::~SolutionDataStorage()
  {
    clear();
  }

  template<typename T>
  void SolutionDataStorage<T>::push(T *solution,
				    double timestamp)
  {
    FUNCNAME("SolutionDataStorage<T>::push()");

    // If pop was the last operation, cleanup and reset the data storage.
    if (poped) {
      clear();
      poped = false;
    }
    
    std::vector<FiniteElemSpace*> feSpace(solution->getFESpaces().size());
    for (int i = 0; i < feSpace.size(); i++) {

      int found = -1;
      for (int j = 0; j < i; j++) {
	if (solution->getFESpace(j) == solution->getFESpace(i)) {
	  found = j;
	  break;
	}
      }

      if (found == -1) {
	feSpace[i] = new FiniteElemSpace();
	*(feSpace[i]) = *(solution->getFESpace(i));       	    
	memoryUsage += feSpace[i]->calcMemoryUsage();
      } else {
	feSpace[i] = feSpace[found];
      }	  
    }

    SystemVector *vec = new SystemVector("tmp", feSpace, solution->getFESpaces().size());
    vec->createNewDOFVectors("tmp");
    vec->setCoarsenOperation(COARSE_INTERPOL);
    vec->interpol(solution, 1.0);
    memoryUsage += vec->calcMemoryUsage();

    solutions.push_back(new SolutionData<T>(vec, feSpace, timestamp, idcounter++));

    lastPos++;

    /*
    
    if (memoryUsage / (1024 * 1024) > maxMemoryUsage) {
      for (int i = 0; i < solutions.size() - 1; i++) {
	if (solutions[i]->isSerialized() == false) {
	  std::ostringstream oss;
	  oss << writeDirectory << "/solutiondata_" << i << ".ser";
	  solutions[i]->serialize(oss.str());
	}
      }
    } 
    */  
  }

  template<typename T>
  bool SolutionDataStorage<T>::pop(T **solution,
				   double *timestamp)
  {
    if (lastPos < 0) {
      return false;
    }

    solutions[lastPos]->getData(solution, timestamp);

    lastPos--;
    poped = true;

    return true;
  }

  template<typename T>
  void SolutionDataStorage<T>::clear()
  {
    for (int i = 0; i < static_cast<int>(solutions.size()); i++) {
      delete solutions[i];
    }

    solutions.clear();

    lastPos = -1;
    memoryUsage = 0;
  }

  template<typename T>	  
  void SolutionDataStorage<T>::addContainer(std::string name)
  {
    std::vector<WorldVector<double> > emptyVec;
    containers.insert(std::pair<std::string, std::vector<WorldVector<double> > >(name, emptyVec) );
    containersPos.insert(std::pair<std::string, int>(name, -1));
  }

  template<typename T>	  
  void SolutionDataStorage<T>::push(std::string name, WorldVector<double> value)
  {
    containers[name].push_back(value);
    containersPos[name] = containers[name].size() - 1;
  }

  template<typename T>	  
  void SolutionDataStorage<T>::pop(std::string name, WorldVector<double> &value)
  {
    value = containers[name][containersPos[name]];
    containersPos[name]--;
  }

  template<typename T>
  void SolutionDataStorage<T>::reset(std::string name)
  {
    containersPos[name] = containers[name].size() - 1;
  }

  template<typename T>
  void SolutionDataStorage<T>::clear(std::string name)
  {
    containers[name].clear();
    containersPos[name] = -1;
  }

}
