// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file parareal/ProblemBase.h */

#ifndef AMDIS_PARAREAL_PROBLEMBASE_H
#define AMDIS_PARAREAL_PROBLEMBASE_H

#include <vector>
#include "CreatorInterface.h"
#include "DOFVector.h"
#include "SystemVector.h"
#include "ProblemInstat.h"

namespace AMDiS {

  template<typename T>
  class ParaRealProblemBase
  {
  public:
    ParaRealProblemBase(CreatorInterface<T> *creator) 
      : storeFineSolutions(false),
	storeCoarseSolutions(false),
	vectorCreator(creator)
    {
      fineSolutions.clear();
      coarseSolutions.clear();
    }
    
    virtual ~ParaRealProblemBase() {};

    void storeSolution(T *vec)
    {
      if (storeFineSolutions) {
	storeFineSolution(vec);
      }
      if (storeCoarseSolutions) {
	storeCoarseSolution(vec);
      }
    }

    void storeFineSolution(T *vec)
    {
      T *tmp = vectorCreator->create();
      *tmp = *vec;
      fineSolutions.push_back(tmp);
    }

    void storeCoarseSolution(T *vec)
    {
      T *tmp = vectorCreator->create();
      *tmp = *vec;
      coarseSolutions.push_back(tmp);
    }

    void setStoreSolutions(bool fine, bool coarse) {
      storeFineSolutions = fine;
      storeCoarseSolutions = coarse;
    }

    void setInitSolutionVec(T *vec) {
      initSolutionVec = vec;
    }

    void setStoreInitSolution(bool store) {
      storeInitSolution = store;
    }

    std::vector<T*> *getFineSolutions() {
      return &fineSolutions;
    }
    
    std::vector<T*> *getCoarseSolutions() {
      return &coarseSolutions;
    }

    void clearFineSolutions() {
      fineSolutions.clear();
    }

    void clearCoarseSolutions() {
      coarseSolutions.clear();
    }

    /** \brief
     * This functions is called between switching from coarse
     * time grid to fine time grid, or vice versa. Can be overwritten
     * to clean up some data.
     */
     virtual void cleanUp() {};
    
  protected:
    /** \brief
     * Stores all solutions on the fine time grid.
     */
    std::vector<T*> fineSolutions;
    
    /** \brief
     * Stores all solutions on the coarse time grid.
     */
    std::vector<T*> coarseSolutions;
    
    /** \brief
     * If true, after closeTimestep() the solution is stored
     * in vector \ref fineSolutions .
     */
    bool storeFineSolutions;
    
    /** \brief
     * If true, after closeTimestep() the solution is stored
     * in vector \ref coarseSolution .
     */
    bool storeCoarseSolutions;
    
    /** \brief
     * If true, the initial solution is also stored in either
     * \ref fineSolutions or in \ref coarseSolutions .
     */
    bool storeInitSolution;
    
    /** \brief
     * Creator interface to create new vectors (either DOFVector or
     * SystemVector).
     */
    CreatorInterface<T> *vectorCreator;
    
    /** \brief
     * The initial solution for the problem can be set to this
     * vector.
     */
    T* initSolutionVec;
  };



  class ParaRealScal : public ProblemInstatScal,
		       public ParaRealProblemBase< DOFVector<double> >
  {
  public:
    ParaRealScal(char* name,
		 ProblemScal *prob,
		 ProblemStatBase *initialProb = NULL)
      : ProblemInstatScal(name, prob, initialProb),
	ParaRealProblemBase< DOFVector<double> >(new DOFVector<double>::Creator(prob->getFESpace()))
    {}
    
    virtual void closeTimestep(AdaptInfo *adaptInfo) {
      ProblemInstatScal::closeTimestep(adaptInfo);
      
      storeSolution(problemStat->getSolution());
    }
    
    virtual void solveInitialProblem(AdaptInfo *adaptInfo) {
      if (initSolutionVec == NULL) {
	ProblemInstatScal::solveInitialProblem(adaptInfo);
      } else {
	*(problemStat->getSolution()) = *initSolutionVec;
      }
      
      if (storeInitSolution) {
	storeSolution(problemStat->getSolution());
      }
    }   
  };


  class ParaRealVec : public ProblemInstatVec,
		      public ParaRealProblemBase< SystemVector >
  {
  public:
    ParaRealVec(char *name,
		ProblemVec *prob,
		ProblemStatBase *initialProb = NULL)
      : ProblemInstatVec(name, prob, initialProb),
	ParaRealProblemBase< SystemVector >(new SystemVector::Creator(name, 
								      prob->getFESpaces(),
								      prob->getNumComponents()))
    {}

    virtual void closeTimestep(AdaptInfo *adaptInfo) {
      ProblemInstatVec::closeTimestep(adaptInfo);
      
      storeSolution(problemStat->getSolution());
    }
    
    virtual void solveInitialProblem(AdaptInfo *adaptInfo) {
      if (initSolutionVec == NULL) {
	ProblemInstatVec::solveInitialProblem(adaptInfo);
      } else {
	*(problemStat->getSolution()) = *initSolutionVec;
      }
      
      if (storeInitSolution) {
	storeSolution(problemStat->getSolution());
      }
    }   
  };
}

#endif
